# -*- coding:utf-8 -*-

"""
Adapted with permission from ReportLab's DocEngine framework
"""


import json, logging

from django import forms
from django.conf import settings
from django.contrib import messages
from django.contrib.staticfiles.storage import staticfiles_storage
from django.forms.utils import flatatt, ErrorList
from django.utils.encoding import force_text
from django.utils.html import format_html, format_html_join
from django.utils.safestring import mark_safe

logger = logging.getLogger('default')

class FileForm(forms.Form):
    file = forms.FileField(required=True)

class LabelledFileForm(forms.Form):
    file = forms.FileField(required=True)
    label = forms.CharField()


class TextCharInput(forms.widgets.TextInput):

    def render(self, name, value, attrs=None):
        if value is None:
            value = 'Click to edit'
        final_attrs = self.build_attrs(attrs, type=self.input_type, name=name)
        return format_html('<span{0}>{1}</span>', flatatt(final_attrs), force_text(self._format_value(value)))


# https://djangosnippets.org/snippets/2631/
def add_control_label(f):
    'add class to label_tag'
    def control_label_tag(self, contents=None, attrs=None):
        if attrs is None:
            attrs = {}
        attrs['class'] = 'control-label'
        return f(self, contents, attrs)
    return control_label_tag


class BSErrorList(ErrorList):

    def as_div(self):
        if not self: return ''
        return format_html('<div class="alert {0}">{1}</div>',
                    settings.MESSAGE_TAGS[messages.ERROR],
                    format_html_join('', '<p>{0}</p>',
                        ((force_text(e),) for e in self)
                    )
          )


class ReadOnlyMixin(object):

    class NewMeta:
        readonly = tuple()

    def __init__(self, *args, **kwargs):
        super(ReadOnlyMixin, self).__init__(*args, **kwargs)
        readonly = self.NewMeta.readonly
        if not readonly:
            return
        for name, field in list(self.fields.items()):
            if name in readonly:
                field.widget = SpanWidget(attrs=getattr(field.widget, 'attrs', {}))
            elif not isinstance(field, SpanField):
                continue
            field.widget.original_value = str(getattr(self.instance, name))


class Html5Mixin(object):
    '''
    Implements HTML5 widgets
    Currently focused on date field
    '''
    class Media:
        js = (
            staticfiles_storage.url("common/js/html5form.js"),
            )

    def __init__(self, *args, **kwargs):
        super(Html5Mixin, self).__init__(*args, **kwargs)
        self.Media.js += (staticfiles_storage.url("common/js/html5form.js"),)
        for name, fld in list(self.fields.items()):
            _x = fld.widget.__class__.__name__
            if _x in ('DateInput', 'TimeInput'):
                _x = _x.replace('Input','')
                fld.widget.input_type = _x.lower()
            elif _x == 'DateTimeInput':
                fld.widget.input_type = 'datetime-local'
            if _x=='Date':
                fld.widget.format="%Y-%m-%d"


class Html5Form(Html5Mixin, forms.Form):
    pass


class Html5ModelForm(Html5Mixin, forms.ModelForm):
    pass


class BootstrapMixin(object):
    """
    Inherit with forms.Form or forms.ModelForm
    For example:
        class AnyForm(forms.Form, BootstrapForm):
            pass

        class AnyModelForm(forms.ModelForm, BootstrapForm):
            pass
    """
    bs_horizontal = True
    error_class = BSErrorList


    def render_errors(self):
        if not self.errors:
            return ""
        output = []
        output.append('<div class="alert alert-danger error">')
        output.append('<button type="button" class="close" data-dismiss="alert" aria-hidden="true">&times;</button>')
        output.append('<p><strong>%s</strong></p><ul>' % 'There are errors while submitting the form:')
        for field, error in list(self.errors.items()):
            output.append('<li><strong>%s</strong> %s</li>' % (field.title(), error[0]))
        output.append('</ul></div>')
        return mark_safe('\n'.join(output))

    def __init__(self, *args, **kwargs):
        super(BootstrapMixin, self).__init__(*args, **kwargs)
        self.error_class = BSErrorList
        for name, fld in list(self.fields.items()):
            if not fld.show_hidden_initial:
                if fld.widget.__class__.__name__ == "RadioSelect":
                    #FIXME: we might be missing other custom attributes enforced upstream
                    fld.widget = BootstrapRadioSelect(choices=fld.choices, inline=False)
                else:
                    _class = fld.widget.attrs.get('class','')
                    if len(_class)>0:
                        _class+=' '
                    if fld.widget.__class__.__name__ in \
                            ('NullBooleanSelect', 'BooleanSelect', 'CheckboxInput'):
                        _class.replace('form-control','')
                    else:
                        if 'form-control' not in _class:
                            _class+='form-control'
                    fld.widget.attrs.update({'class':_class,})

    def _render_field(self, obj, level=0, siblings=0):
        if getattr(obj, 'is_fieldset', False):
            output = '<fieldset'
            if getattr(obj, 'css_classes', False):
                output+=' class="%s"' % obj.css_classes
            output += '>'
            for flds in obj:
                kw = {'level': level+1,}
                if getattr(flds, 'is_fieldset', False):
                    kw['siblings'] = len(flds.fieldset.fields)
                else:
                    kw['siblings'] = siblings
                output+=self._render_field(flds,**kw)
            output+='</fieldset>'
            return output
        else:
            if getattr(obj, 'is_hidden', False):
                return  obj.as_widget()

            _div_class = "form-group "
            _label_attrs = {'class': 'control-label',}
            _field_html = '%s'
            _help_tag = '<div class="form-control-static help_block text-muted">%s</div>'

            horizontal_set = (siblings > 1) and (level > 0)

            if horizontal_set:
                _div_class += " col-sm-%d" % (12 // siblings) # bootstrap grid
            else:
                if getattr(obj, 'help_text', False):
                    _field_col_width = 7
                else:
                    _field_col_width = 10
                _div_class += "row"
                #if obj.field.__class__.__name__ == 'GridFormField':
                #    _label_attrs['style']='float:none'
                _label_attrs['class']+= ' col-sm-2'
                _field_html = '<div class="col-sm-%d">' % _field_col_width + '%s</div>'
                _help_tag = '<div class="col-sm-3 form-control-static help_block text-muted">%s</div>'

            row_template = '''
<div class="%(div_class)s">
    %(label)s
    %(field)s
    %(help_text)s
    %(errors)s
</div>
'''
            # check if it is an actual field and not a fieldset from betterforms
            widget_attrs = obj.field.widget.attrs
            widget_classes = widget_attrs.get('class','').split()
            if 'form-control' not in widget_classes and \
                    obj.field.widget.__class__.__name__ not in \
                    ('NullBooleanSelect', 'BooleanSelect',
                     'CheckboxInput','RadioSelect','BootstrapRadioSelect'):
               widget_attrs.update({'class': '%s form-control' % ' '.join(widget_classes)})
            if not obj.field.required:
                _label_attrs['class'] += " text-muted"

            row_dict = {
                "div_class" : _div_class,
                "field" :  _field_html % obj.as_widget(),
                "label" : obj.label_tag(attrs=_label_attrs),
                "help_text" : _help_tag % obj.help_text or '',
                "errors": obj.errors,
            }
            if obj.errors:
                row_dict['help_text']='';
                row_dict["div_class"] = "row form-group alert alert-danger"
                #obj.field.widget.attrs["class"]="error"

            return row_template % row_dict

    def as_div(self):
        'This wont be used by betterforms'
        # FIXME: not integrating with betterforms
        #if BETTERFORMS_LOADED and isinstance(self, BetterForm):
        #    return self
        output = []
        if self.bs_horizontal:
            self._field_html = '<div class="col-sm-%d">%s</div>'
            self._label_attrs = {'class':'col-sm-2 control-label',}
            self._help_tag = '<div class="col-sm-3 help_block">%s</div>'
        else:
            self._field_html = '<!--%s-->%s'
            self._label_attrs = {}
            self._help_tag = '<div class="help_block">%s</div>'

        for flds in self: #see original Form class __iter__ method
            output.append(self._render_field(flds))

        return mark_safe('\n'.join(output))


class BootstrapForm(forms.Form, BootstrapMixin):
    pass


class BootstrapModelForm(forms.ModelForm):

    def as_div(self):
        "Returns this form rendered as Bootstrap HTML"
        return forms.BaseModelForm._html_output(self,
            normal_row = '''<div class="form-group">%(errors)s%(label_tag)s
%(field)s''',
            error_row = '<li>%s</li>',
            row_ender = '</div>',
            help_text_html = ' <div class="help_block">%s</div>',
            errors_on_separate_row = False)



class JEditableMixin(object):

    form_instance_id = None
    form_object_name = None

    class Media:
        js = (
            staticfiles_storage.url("common/js/ajax_setup.js"),
            staticfiles_storage.url("common/js/jquery.jeditable.js"),
            staticfiles_storage.url("common/js/editable.js"),
            staticfiles_storage.url("common/js/jquery.autosize-min.js"),
            )
        if getattr(settings, 'TRACK_AJAX_CALLS', False):
            js+=('common/js/jquery.ajaxtrack.js',)
        css = {
            'screen': ( staticfiles_storage.url("common/css/editable.css"),
                ),
            }

    def as_div(self):
        "Returns this form rendered as HTML <dt /><dd />s -- excluding the <dl></dl>."
        return forms.BaseModelForm._html_output(self,
            normal_row = '%(errors)s<dt>%(label)s</dt><dd%(html_class_attr)s>%(field)s</dd>%(help_text)s',
            error_row = '<li>%s</li>',
            row_ender = '</dd>',
            help_text_html = ' <span class="helptext">%s</span>',
            errors_on_separate_row = False)

    def generate_jeditable_attrs(self, fieldname):
        _instance = self.instance
        _bf, containingmodel, direct, m2m = _instance._meta.get_field_by_name(fieldname)
        _uce = getattr(self.instance, 'user_can_edit', False)
        attrs = {
            #'field_name': _bf.attname,
            # anyone will be able to edit for now
            'data-can_edit': str(_uce(attr=fieldname)) if _uce else "True",
            'data-instance_name': self.__class__.__name__.lower(),
            'data-instance_id': self.instance.pk,
            'data-field_name': _bf.name,
            }
        if getattr(_bf, 'max_length', False):
            attrs.update({'maxlength': getattr(_bf, 'max_length')})
        _klss = ''
        if _bf.choices:
            _klss= 'choice '
            #attrs['choices'] = json.dumps(dict(getattr(_bf, 'choices', {})))
        _klss+=_bf.__class__.__name__.lower()
        attrs['class'] = "edit_%s" % _klss
        if getattr(_bf, 'extra_attrs', False):
            attrs['data-mce-config'] = json.dumps(getattr(_bf, 'extra_attrs', {}))
        #attrs = JEditableAttrsDict(attrs)
        return attrs

    def __init__(self, *args, **kwargs):
        super(JEditableMixin, self).__init__(*args, **kwargs)
        _cls_names = []
        for fld in self.fields:
            _form_field = self.fields[fld]
            if _form_field.widget.__class__.__name__ in (
                     'BootstrapRadioSelect', 'RadioSelect'):
                _form_field.widget = forms.Select(choices=_form_field.choices)
            _cls_name = _form_field.__class__.__name__
            if _cls_name not in _cls_names:
                _cls_names.append(_cls_name)
            if 'DateField' == _cls_name:
                _cls_names.append('datepicker')
                self.fields[fld].widget = forms.widgets.TextInput(attrs={'class': _cls_names, 'type':'date'})
            if not _form_field.show_hidden_initial:
                attrs = _form_field.widget.attrs
                widget_attrs = self.generate_jeditable_attrs(fld)
                attrs.update(widget_attrs)

                #if 'DateField' == _cls_name:
                #   attrs.update({'data-date':'2014-01-01'})
        # add media as required
        if not getattr(self.Media, 'js', False):
            self.Media.js = ()
            self.Media.css = {'screen': (),}
        if 'BooleanField' in _cls_names or 'NullBooleanField' in _cls_names:
            self.Media.js += (staticfiles_storage.url('common/js/jquery.icheck.min.js'),)
            self.Media.css['screen'] += (staticfiles_storage.url('common/css/icheck/skins/flat/blue.css'),)
        if 'ModelMultipleChoiceField' in _cls_names:
            self.Media.js += (staticfiles_storage.url('common/js/chosen.jquery.min.js'),)
            self.Media.css['screen'] += (staticfiles_storage.url('common/css/chosen.min.css'),)
        # prevent multiple media includes
        self.Media.js = tuple(set(self.Media.js))
        self.Media.css['screen'] = tuple(set(self.Media.css['screen']))


class JEditableFormMixin(JEditableMixin, forms.Form):
    pass


class JEditableModelForm(JEditableMixin,forms.ModelForm):
    pass

class CommentForm(BootstrapForm):
    comment = forms.CharField(widget=forms.Textarea)



'''
dummy module to allow rlgendoc to work and acknowledge the app_name and 
at the same time not require another prep file
'''
from accountifie.common.makepdf import rlgendoc

@rlgendoc
def finreport(*args, **kwargs):
    return kwargs




import datetime
from dateutil.parser import parse
from dateutil.rrule import rrule, DAILY,MO,TU,WE,TH,FR

from accountifie.gl.models import ExternalAccount, ExternalBalance
import accountifie.toolkit.utils as utils
import accountifie.query.query_manager

def externalaccounts(qstring={}):
    return [x.gl_account.id for x in ExternalAccount.objects.all().select_related('gl_account')]


def external_bals_history(qstring={}):
    dt = parse(qstring.get('date', datetime.datetime.now().isoformat())).date()
    company_id = qstring.get('company_id', utils.get_default_company())
    gl_strategy = qstring.get('gl_strategy', None)
    
    acct = qstring.get('acct',None)
    dts = list(rrule(DAILY, dtstart=dt + datetime.timedelta(days=-7), until=dt, byweekday=(MO,TU,WE,TH,FR)))
    dts = [dt.date() for dt in dts]

    external_balances = dict((x.date.isoformat(), float(x.balance)) for x in ExternalBalance.objects.filter(account__id=acct).filter(date__in=dts))
    int_bals_df = accountifie.query.query_manager.QueryManager(gl_strategy).pd_acct_balances(company_id, dict((dt.isoformat(),dt) for dt in dts), acct_list=[acct])
    if acct in int_bals_df.index:
        internal_balances = int_bals_df.loc[acct].to_dict()
    else:
        internal_balances = {}
    bal_checks = []
    
    for dt in dts:
        external = external_balances.get(dt.isoformat(), 0.0)
        internal = internal_balances.get(dt.isoformat(), 0.0)
        bal_checks.append({'Date': dt, 'Internal': internal, 'External': external, 'Diff': external - internal})
    
    return bal_checks


from django.contrib import admin
from django.http import HttpResponseRedirect
from django.core.urlresolvers import reverse

from accountifie.snapshot.models import GLSnapshot

class GLSnapshotAdmin(admin.ModelAdmin):
    list_display=('closing_date', 'short_desc', 'snapped_at',)

    def response_add(self, request, obj, post_url_continue=None):
        return HttpResponseRedirect(reverse('glsnapshots'))

admin.site.register(GLSnapshot, GLSnapshotAdmin)





import os
from django.conf import settings  # noqa

from celery import Celery

proj = settings.PROJECT_NAME
# set the default Django settings module for the 'celery' program.
os.environ.setdefault('DJANGO_SETTINGS_MODULE', proj)


app = Celery(proj)

# Using a string here means the worker will not have to
# pickle the object when using Windows.
app.config_from_object('django.conf:settings')
app.autodiscover_tasks(lambda: settings.INSTALLED_APPS)


@app.task(bind=True)
def debug_task(self):
    print(('Request: {0!r}'.format(self.request)))


"Re-saves all things which might produce GL transactions."
import os, time
from optparse import make_option
import inspect
import importlib

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction

from accountifie.gl.models import Transaction
from accountifie.gl.bmo import  BusinessModelObject
from accountifie.common.api import api_func

from accountifie.query.query_manager_strategy_factory import QueryManagerStrategyFactory


class Command(BaseCommand):

    """
    option_list = BaseCommand.option_list + (
        make_option('--all',
                action='store_true', 
                help='Full recalc of GL'),
        )
    """ 
    
    def handle(self, *args, **options):
        #faster and less likely to mess stuff up.

        klasses = []
        kl_paths = api_func('environment', 'variable_list', 'BMO_MODULES')

        # find all the BMO classes
        for path in kl_paths:
            for name, kl in inspect.getmembers(importlib.import_module(path), inspect.isclass):
                if BusinessModelObject in kl.__bases__:
                    klasses.append(kl)
        
        with transaction.atomic():
            Transaction.objects.all().delete()
            
            for cmpny in [c['id'] for c in api_func('gl', 'company')]:
                QueryManagerStrategyFactory().erase(cmpny)
            
            print("deleted all transactions")
            QueryManagerStrategyFactory().set_fast_inserts('*', True)
            for klass in klasses:
                print('working on', klass)    
                qs =klass.objects.all()
                for obj in qs:
                    obj.update_gl()
                print('finished with', klass)    
            QueryManagerStrategyFactory().set_fast_inserts('*', False)
            QueryManagerStrategyFactory().take_snapshot('*')
        print("updated %d transactions" % qs.count())



#!/usr/bin/env python

from setuptools import setup, find_packages

setup(name='accountifie',
      description='Core Accouting Functionality',
      author='Ian McAllister',
      version='0.3.4',
      author_email='ian@mcallisternevins.com',
      url='https://github.com/imcallister/accountifie',
      packages=find_packages(),
      include_package_data=True,
      )



"""Date objects all nicely joinable together to allow SQL GROUP BY.

The calendar actually gets constructed by functions in __init__.py
"""
from datetime import date
from django.db import models


class Year(models.Model):
    id = models.IntegerField(primary_key=True, help_text="4 digit year e.g. 2009")
    def __unicode__(self):
        return str(self.id)
        
class Quarter(models.Model):
    id = models.CharField(primary_key=True, max_length=6, help_text='e.g. "2009Q1"')
    year = models.ForeignKey(Year, db_index=True)
    def __unicode__(self):
        return self.id
    
class Month(models.Model):
    id = models.CharField(primary_key=True, max_length=7, help_text='e.g. "2009M04"')
    quarter = models.ForeignKey(Quarter, db_index=True)
    year = models.ForeignKey(Year, db_index=True)
    def __unicode__(self):
        return str(self.id)
        
    def first_of_month(self):
        yyyy, mm = self.id.split('M')
        return date(int(yyyy), int(mm), 1)
    
    def day_of_month(self, day):
        yyyy, mm = self.id.split('M')
        return date(int(yyyy), int(mm), day)

class Week(models.Model):
    "Identified by YYYYWNN  e.g. 2007W29"
    id = models.CharField(primary_key=True, max_length=7)
    first_day = models.DateField(null=True)  #can be last year, hard to initialize in a new system
    last_day = models.DateField(null=True) #can be next year, hard to initialize at end of year
    start_month = models.ForeignKey(Month, related_name='start_week_set')
    end_month = models.ForeignKey(Month, related_name='end_week_set')
    year = models.ForeignKey(Year)
    def __unicode__(self):
        return str(self.id)

class Day(models.Model):
    id = models.DateField(primary_key=True)
    month = models.ForeignKey(Month, db_index=True)
    quarter = models.ForeignKey(Quarter, db_index=True)
    year = models.ForeignKey(Year, db_index=True)
    def __unicode__(self):
        return '%04d-%02d-%02d' % (self.id.year, self.id.month, self.id.day)
    



import datetime
import copy
from decimal import Decimal
from dateutil.parser import parse
import logging

from accountifie.reporting.models import Report
from accountifie.toolkit.utils import DZERO
import accountifie.toolkit.utils as utils
from accountifie.common.api import api_func
import accountifie.reporting.rptutils as rptutils


logger = logging.getLogger('default')

class TrialBalance(Report):
    """Left/Right display of each active account on the given date.

    Shamelessly piggbacking on the Report class without really inheriting
    anything, so that the find_reports function will find it.
    """

    def __init__(self, company_id, date=None):
        config = {'description': 'Trial Balance',
                  'calc_type': 'as_of',
                  'date': date}
        
        super(TrialBalance, self).__init__(company_id, **config)

        self.label_map = None
        self.link_map = lambda x: utils.acct_history_link(x.name)
        self.works_for = [cmpny['id'] for cmpny in api_func('gl', 'company')]
        self.columns = {'Debits':'Debits', 'Credits': 'Credits'}
        self.column_order = ['Debits', 'Credits']
        
        
    def set_columns(self, columns, column_order=None):
        # columns are fixed... don't mistakenly adjust to dates
        pass

    def configure(self, config, path=None):
        qs_matches = rptutils.qs_parse(config)
        if len(qs_matches) == 0:
            raise ValueError('Unexpected query string: %s' % repr(config))
        elif len(qs_matches) > 1:
            raise ValueError('Unexpected query string: %s' % repr(config))
        else:
            config['config_type'] = qs_matches[0]


        if config['config_type'] == 'shortcut':
            config.pop('config_type')
            config.update(rptutils.parse_shortcut(config['col_tag']))

        if config['config_type'] == 'date':
            dt = rptutils.date_from_shortcut(config['date'])
            config.update(rptutils.config_fromdate(self.calc_type, self.description, dt))
        elif config['config_type'] == 'date_range':
            dt = rptutils.date_from_shortcut(config['to'])
            config.update(rptutils.config_fromdate(self.calc_type, self.description, dt))
        else:
            raise ValueError('Unexpected query string: %s' % repr(config))
        
        self.set_columns(config['columns'], column_order=config.get('column_order'))
        self.date = config.get('date')
        self.path = config.get('path')
        self.title = 'Trial Balance as of %s' % self.date.strftime('%d-%b-%y')


    def calcs(self):
        bals = self.query_manager.pd_acct_balances(self.company_id, {'balance': self.date})
        
        accts = api_func('gl', 'account')
        accts_map = dict((a['id'], a) for a in accts)

        # dataframe.apply won't work until v0.17 for lambda returning dict... so have to do in roundabout way till then
        debits_map = lambda x: x['balance'] if accts_map[x.name]['role'] in ['asset', 'expense'] else 0
        credits_map = lambda x: -x['balance'] if accts_map[x.name]['role'] in ['liability', 'income', 'capital'] else 0
        
        bals['Debits'] = bals.apply(debits_map, axis=1)
        bals['Credits'] = bals.apply(credits_map, axis=1)
        del bals['balance']

        bals['fmt_tag'] = 'item'
        label_map = lambda x: x + ': ' + accts_map[x]['display_name']
        bals['label'] = bals.index.map(label_map)
        
        totals = bals[self.column_order].sum(axis=0)
        totals['fmt_tag'] = 'major_total'
        totals['label'] = 'Totals'

        bals.loc['Total'] = totals

        bals['index'] = bals.index
        data = bals.to_dict(orient='records')

        credits_link_map = lambda x: utils.acct_history_link(x['index']) if x['index'] != 'Total' and x['Credits'] != '' else ''
        debits_link_map = lambda x: utils.acct_history_link(x['index']) if x['index'] != 'Total' and x['Debits'] != '' else ''
        
        for row in data:
            row['Credits'] = {'text': row['Credits'], 'link':credits_link_map(row)}
            row['Debits'] = {'text': row['Debits'], 'link': debits_link_map(row)}
        
        return data


import os
import csv
import json
from ast import literal_eval
from io import StringIO
import datetime
import gzip
import csv
import datetime
from dateutil.parser import parse

import logging

from django.contrib import messages
from django.contrib.auth.decorators import login_required, user_passes_test
from django.core.management import call_command
from django.http import HttpResponseRedirect, HttpResponse


from accountifie.gl.models import Company
from . import utils

logger = logging.getLogger('default')


def company_context(request):
    """Context processor referenced in settings.
    This puts the current company ID into any request context, 
    thus allowing templates to refer to it

    This is not a view.
    """

    company_id = utils.get_company(request)
    data = {'company_id': company_id}
    if company_id:
        try:
            company = Company.objects.get(pk=company_id)
            data.update({'company': company})
        except Company.DoesNotExist:
            pass
    return data


@login_required
def choose_company(request, company_id):
    "Hit this to switch companies"

    company_list = [x.id for x in Company.objects.all()]

    if company_id not in company_list:
        raise ValueError('%s is not a valid company' % company_id)
    request.session['company_id'] = company_id
    dest = request.META.get('HTTP_REFERER', '/')
    return HttpResponseRedirect(dest)



@user_passes_test(lambda u: u.is_superuser)
def dump_fixtures(request):
    output = StringIO()

    fixture = request.GET.get('fixture', None)

    try:
        if fixture:
            call_command('dumpdata', fixture, '--indent=2', stdout=output)
        else:
            call_command('dumpdata', '--indent=2', stdout=output)

        data = output.getvalue()
        output.close()

        if fixture:
            file_label = 'fixtures_%s_%s' % (fixture, datetime.datetime.now().strftime('%d-%b-%Y_%H-%M'))
        else:
            file_label = 'fixtures_all_%s' % datetime.datetime.now().strftime('%d-%b-%Y_%H-%M')
        response = HttpResponse(data, content_type="application/json")
        response['Content-Disposition'] = 'attachment; filename=%s' % file_label
        return response
    except:
        dest =  request.META.get('HTTP_REFERER', '/')
        messages.info(request, 'Fixture name not recognized: %s' % fixture)
        return HttpResponseRedirect(dest)



import six
from deepdiff import DeepDiff

#import accountifie.gl.bmo
from django.conf import settings
from django.db import models
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core import urlresolvers



def get_child_paths(path):
    path_len = len(path.split('.'))
    all_sub_paths = [x.path for x in Account.objects.filter(path__contains=path) if len(x.path.split('.')) > path_len]
    all_child_paths = ['.'.join(x.split('.')[:(path_len+1)]) for x in all_sub_paths]
    return list(set(all_child_paths))


class Account(models.Model):
    id = models.CharField(max_length=20, primary_key=True, 
        help_text="Unique short code or number")
    path = models.CharField(max_length=100, db_index=True, unique=True,
        help_text="dotted-path reference to show location on balance sheet")
    ordering = models.IntegerField(default=0)
    display_name = models.CharField(max_length=100, blank=True,
        help_text="the name you would use on a report title")
    role = models.CharField(max_length=10,
        help_text="used in P&L queries and for picklists",
        choices = [
            ('asset', 'asset'),
            ('liability', 'liability'),
            ('income', 'income'),
            ('expense', 'expense'),
            ('capital', 'capital'),
            ]
        )
        
    class Meta:
        ordering = ('id',)
        app_label = 'gl'
        db_table = 'gl_account'
    
    def __str__(self):
        return '%s: %s' % (self.id, self.display_name)

"""
class Transaction(models.Model, accountifie.gl.bmo.BusinessModelObject):
    company = models.ForeignKey('gl.Company')
    date = models.DateField(db_index=True)
    date_end = models.DateField(db_index=True, blank=True, null=True)
    comment = models.CharField(max_length=100)
    long_desc = models.CharField(max_length=200, blank=True, null=True)
    bmo_id = models.CharField(max_length=100)
    content_type = models.ForeignKey(ContentType)
    object_id = models.PositiveIntegerField()
    source_object = GenericForeignKey()

    class Meta:
        app_label = 'gl'
        db_table = 'gl_transaction'

    def __str__(self):
        return '%s' % self.comment

    def save(self):
        db_tr = Transaction.objects.filter(bmo_id=self.bmo_id).first()
        if db_tr is None:
            super(Transaction, self).save()
        else:
            if len(DeepDiff(self._to_dict(), db_tr._to_dict())) > 0:
                print('CHANGED')
                print(DeepDiff(self, db_tr))
                super(Transaction, self).save()
            else:
                print('not saving it')
        self.update_gl()

    def delete(self):
        tranlines = TranLine.objects.filter(transaction__id=self.id)
        for tl in tranlines:
            tl.delete()
        models.Model.delete(self)

    def get_admin_url(self):
        #https://djangosnippets.org/snippets/1916/
        content_type = ContentType.objects.get_for_model(self.__class__)
        return urlresolvers.reverse("admin:%s_%s_change" % (content_type.app_label, content_type.model), args=(self.id,))

    def _to_dict(self):
        return dict((f.attname, getattr(self, f.attname))
                    for f in self._meta.fields if f.attname not in ['id'])

    def _tlines_to_dict(self):
        return dict((tl.id, tl._to_dict()) for tl in self.tranline_set.all())

    def get_changes(self):
        db_tr = Transaction.objects.filter(bmo_id=self.bmo_id).first()
        if db_tr is None:
            return {}
        return DeepDiff(self, db_tr)
"""


class TranLine(models.Model):
    #transaction = models.ForeignKey(Transaction)
    company = models.ForeignKey('gl.Company', blank=True, null=True)
    date = models.DateField(db_index=True)
    date_end = models.DateField(db_index=True, blank=True, null=True)
    comment = models.CharField(max_length=100, blank=True, null=True)
    account = models.ForeignKey(Account, blank=True, null=True)
    amount = models.DecimalField(max_digits=11, decimal_places=2)
    counterparty = models.ForeignKey('gl.Counterparty', blank=True, null=True)
    tags = models.CharField(max_length=200, blank=True)
    bmo_id = models.CharField(max_length=100, blank=True, null=True)
    content_type = models.ForeignKey(ContentType, blank=True, null=True)
    object_id = models.PositiveIntegerField(blank=True, null=True)
    source_object = GenericForeignKey()

    class Meta:
        app_label = 'gl'
        db_table = 'gl_tranline'

    def __str__(self):
        return '%.2f: Account %s' %(self.amount, self.account)

    @property
    def date(self):
        return self.transaction.date

    def _to_dict(self):
        return dict((f.attname, getattr(self, f.attname))
                    for f in self._meta.fields if f.attname not in ['id'])


class ExternalBalance(models.Model):
    "Assertion that account X has balance Y on date Z"
    company = models.ForeignKey('gl.Company')
    account = models.ForeignKey(Account)
    date = models.DateField(db_index=True)
    balance = models.DecimalField(max_digits=11, decimal_places=2)
    comment = models.CharField(max_length=100,blank=True,null=True)

    class Meta:
        app_label = 'gl'
        db_table = 'gl_externalbalance'

class DepreciationPolicy(models.Model):
    cap_account = models.ForeignKey(Account, related_name='deppolicy_acct')
    depreciation_account = models.ForeignKey(Account, related_name='deppolicy_depacct')
    depreciation_period = models.PositiveIntegerField()

    class Meta:
        app_label = 'gl'
        db_table = 'gl_depreciationpolicy'
    


import datetime
import pandas as pd
from dateutil.parser import parse


from accountifie.reporting.models import Report
import accountifie.gl.models as gl
import accountifie.toolkit.utils as utils
from accountifie.common.api import api_func


class AccountActivity(Report):

    def __init__(self, company_id, date=None):
        config = {'description': 'Account Activity',
                  'calc_type': 'as_of',
                  'date': date}

        super(AccountActivity, self).__init__(company_id, **config)

        self.label_map = None
        self.link_map = lambda x: utils.acct_history_link(x.name)
        self.works_for = [cmpny['id'] for cmpny in api_func('gl', 'company')]
        self.acct_list = None
        self.equity_sign = False




    def calcs(self):
        if self.path:
            bals = self.query_manager.pd_acct_balances(self.company_id,self.columns, paths=[self.path]).fillna(0.0)
        elif self.acct_list:
            bals = self.query_manager.pd_acct_balances(self.company_id,self.columns, acct_list=self.acct_list, excl_contra=['4150']).fillna(0.0)
        else:
            bals = self.query_manager.pd_acct_balances(self.company_id,self.columns).fillna(0.0)


        if self.equity_sign:
            bals = bals * (-1.0)

        bals.loc['Total'] = bals.apply(sum, axis=0)
        accts = gl.Account.objects.all()
        acct_map = dict((a.id, a.display_name) for a in accts)
        label_map = lambda x: x + ': ' + acct_map[x] if x in acct_map else x
        link_map = lambda x: utils.acct_history_link(x['index']) if x['index'] != 'Total' else ''

        bals['fmt_tag'] = 'item'
        bals['label'] = bals.index.map(label_map)
        bals['index'] = bals.index

        if 'Change' in bals.columns:
            is_small = lambda row: (abs(row['Change']) < 0.5 and row['index'] != 'Total')
            bals = bals[~bals.apply(is_small, axis=1)]

        bals.loc['Total', 'fmt_tag'] = 'major_total'

        data = bals.to_dict(orient='records')

        for row in data:
            for col in self.column_order:
                row[col] = {'text': row[col], 'link': link_map(row)}

        return data



from django.template import Context
from django.template.loader import get_template


def basic_modal(content, modal_title, modal_id):
    context = {}
    context['content'] = content
    context['modal_id'] = modal_id
    context['modal_title'] = modal_title

    tmpl = get_template('modal_cmpnt.html')
    return tmpl.render(Context(context))



from django.conf import settings
from django import http
from django.core.mail import mail_admins
from django.shortcuts import render
from django.template import RequestContext
import traceback

import sys

# Temporary, from http://code.djangoproject.com/attachment/ticket/6094/6094.2008-02-01.diff
from django.core.urlresolvers import RegexURLResolver
def resolver(request):
    """
    Returns a RegexURLResolver for the request's urlconf.

    If the request does not have a urlconf object, then the default of
    settings.ROOT_URLCONF is used.
    """
    from django.conf import settings
    urlconf = getattr(request, "urlconf", settings.ROOT_URLCONF)
    return RegexURLResolver(r'^/', urlconf)


class StandardExceptionMiddleware(object):
    def process_exception(self, request, exception):
        # Get the exception info now, in case another exception is thrown later.

        if isinstance(exception, http.Http404):
            return self.handle_404(request, exception)
        else:
            return self.output_exception(request, exception)

    def output_exception(self, request, exception):
        context = {}
        context['exception_detail'] = sys.exc_info()[1]
        context['traceback'] = traceback.format_exc().split('\n')
        return render(request, 'exception.html', context)

    def handle_404(self, request, exception):
        if settings.DEBUG:
            from django.views import debug
            return debug.technical_404_response(request, exception)
        else:
            callback, param_dict = resolver(request).resolve404()
            return callback(request, **param_dict)


    def handle_500(self, request, exception):
        exc_info = sys.exc_info()
        if settings.DEBUG:
            return self.debug_500_response(request, exception, exc_info)
        else:
            self.log_exception(request, exception, exc_info)
            return self.production_500_response(request, exception, exc_info)


    def debug_500_response(self, request, exception, exc_info):
        from django.views import debug
        return debug.technical_500_response(request, *exc_info)


    def production_500_response(self, request, exception, exc_info):
        '''Return an HttpResponse that displays a friendly error message.'''
        callback, param_dict = resolver(request).resolve500()
        return callback(request, **param_dict)


    def exception_email(self, request, exc_info):
        subject = 'Error (%s IP): %s' % ((request.META.get('REMOTE_ADDR') in settings.INTERNAL_IPS and 'internal' or 'EXTERNAL'), request.path)
        try:
            request_repr = repr(request)
        except:
            request_repr = "Request repr() unavailable"
        message = "%s\n\n%s" % (_get_traceback(exc_info), request_repr)
        return subject, message


    def log_exception(self, request, exception, exc_info):
        subject, message = self.exception_email(request, exc_info)
        mail_admins(subject, message, fail_silently=True)



def _get_traceback(self, exc_info=None):
    """Helper function to return the traceback as a string"""
    return '\n'.join(traceback.format_exception(*(exc_info or sys.exc_info())))



import accountifie.toolkit.utils.datefuncs as datefuncs
from . import shortcuts

MONTHS = ['Jan','Feb','Mar','Apr','May','Jun','Jul','Aug','Sep','Oct','Nov','Dec']


def get_period_id(config):
    if config['period'] == 'year':
        return str(config['year'])
    elif config['period'] == 'half':
        return '%sH%s' % (config['year'], config['half'])
    elif config['period'] == 'quarter':
        return '%sQ%s' % (config['year'], config['quarter'])
    elif config['period'] == 'month':
        return '%sM%02d' % (config['year'], int(config['month']))
    elif config['period'] == 'day':
        return 'D%d-%02d-%02d' % (int(config['year']), int(config['month']), int(config['day']))
    else:
        raise ValueError("Bad config %s" % repr(config))


def monthly_periods(calc_type, start, end):
    months = list(datefuncs.monthrange(start, end))
    if calc_type == 'diff':
        columns = ['%sM%s' % (x[0], '%02d' % x[1]) for x in months]
        column_titles = ['%s %s' % (MONTHS[x[1]-1], x[0]) for x in months]    
    elif calc_type == 'as_of':
        columns = [datefuncs.end_of_month(x[1], x[0]).isoformat() for x in months]
        column_titles = columns
    else:
        raise ValueError("Bad calc_type %s" % repr(calc_type))
    
    return columns, column_titles

def quarterly_periods(calc_type, start, end):
    quarters = list(datefuncs.quarterrange(start, end))
    if calc_type == 'diff':
        columns = ['%sQ%s' % (x[0], '%0d' % x[1]) for x in quarters]
    elif calc_type == 'as_of':
        columns = [datefuncs.end_of_quarter(x[1], x[0]).isoformat() for x in quarters]
    else:
        raise ValueError("Bad calc_type %s" % repr(calc_type))
    
    column_titles = columns
    return columns, column_titles

def half_periods(calc_type, start, end):
    halfs = list(datefuncs.halfrange(start, end))
    if calc_type == 'diff':
        columns = ['%sH%s' % (x[0], '%0d' % x[1]) for x in halfs]
    elif calc_type == 'as_of':
        columns = [datefuncs.end_of_half(x[1], x[0]).isoformat() for x in halfs]
    else:
        raise ValueError("Bad calc_type %s" % repr(calc_type))
    
    column_titles = columns
    return columns, column_titles

def annual_periods(calc_type, start, end):
    years = list(datefuncs.annualrange(start, end))
    if calc_type == 'diff':
        columns = ['%s' % y  for y in years]
    elif calc_type == 'as_of':
        columns = [datefuncs.end_of_year(x).isoformat() for x in years]
    else:
        raise ValueError("Bad calc_type %s" % repr(calc_type))

    column_titles = columns
    return columns, column_titles

def day_periods(calc_type, start, end):
    columns = [start.isoformat(), end.isoformat()]
    column_titles = ['Previous Day', 'Current Day']
    return columns, column_titles

def gen_periods(calc_type, config):
    period_id = get_period_id(config)
    start = datefuncs.start_of_period(period_id)
    end = datefuncs.end_of_period(period_id)

    # if just as of end of single period... show chg
    if config['period'] == 'day':
        return single_period(config, start, end)

    if config['by'] == config['period'] and calc_type == 'as_of':
        return single_period(config, datefuncs.end_of_prev_period(period_id), end)

    if config['by'] == 'year':
        return annual_periods(calc_type, start, end)
    elif config['by'] == 'half':
        return half_periods(calc_type, start, end)
    elif config['by'] == 'quarter':
        return quarterly_periods(calc_type, start, end)
    elif config['by'] == 'month':
        return monthly_periods(calc_type, start, end)
    elif config['by'] == 'day':
        return day_periods(calc_type, start, end)
    else:
        raise ValueError("Unrecognised config %s" % repr(config))        


def daterange_periods(calc_type, config):
    start = shortcuts.date_from_shortcut(config['from'])
    end = shortcuts.date_from_shortcut(config['to'])
    
    if config['by'] == 'year':
        return annual_periods(calc_type, start, end)
    elif config['by'] == 'half':
        return half_periods(calc_type, start, end)
    elif config['by'] == 'quarter':
        return quarterly_periods(calc_type, start, end)
    elif config['by'] == 'month':
        return monthly_periods(calc_type, start, end)
    else:
        columns = [start.isoformat(), end.isoformat()]
        column_titles = list(columns)
        return columns, column_titles


# SINGLE PERIODS
def single_period(config, start, end):
    period_id = get_period_id(config)
    columns = [start.isoformat(), period_id, end.isoformat()]
    column_titles = [columns[0], 'chg in period', columns[2]]
    return columns, column_titles





from datetime import datetime
from decimal import Decimal

from django.db.models import Model as djangoModel
from django.core.exceptions import ObjectDoesNotExist, MultipleObjectsReturned
from . import exceptions


class AllChoices(object):
    def __contains__(self, value):
        return True


class AlwaysValidValidator(object):
    def validate(self, val):
        return True


class BaseField(object):
    def __init__(self, kwargs):
        self.transform = kwargs.pop('transform', lambda val:val)


class Field(BaseField):
    position = 0

    def __init__(self, **kwargs):
        super(Field, self).__init__(kwargs)
        self.null = kwargs.pop("null", False)
        self.default = kwargs.pop("default", None)
        if self.default and not self.null:
            raise exceptions.FieldError("You cannot provide a default without setting the field as nullable")
        if 'row_num' in kwargs:
            self.position = kwargs.pop('row_num')
        else:
            self.position = Field.position
            Field.position += 1
        if 'match' in kwargs:
            self.match = kwargs.pop('match')
        self.validator = kwargs.pop('validator', AlwaysValidValidator)
        if 'multiple' in kwargs:
            self.has_multiple = kwargs.pop('multiple')
        self.prepare = kwargs.pop('prepare', lambda val:val)
        if 'keys' in kwargs and isinstance(self, ComposedKeyField):
            self.keys = kwargs.pop('keys')
        self.choices = kwargs.pop('choices', AllChoices())
        if len(kwargs) > 0:
            raise ValueError("Arguments %s unexpected" % list(kwargs.keys()))

    def get_transform_method(self, instance):
        """ Look for transform_<field_name>, else look for the transform parameter, else identity method """
        transform_method = "transform_" + getattr(self, "fieldname", self.field_name)
        transform = getattr(instance, transform_method, self.transform)
        return transform

    def get_prep_value(self, value, instance=None):
        try:
            value = self.prepare(value)
            transform = self.get_transform_method(instance)
            value = transform(value)

            if not value and self.null:
                value = self.default
            else:
                value = self.to_python(value)
            if value not in self.choices:
                if not self.null:
                    raise exceptions.ChoiceError("Value \'%s\' does not belong to %s" % (value, self.choices))
                value = None
            if not self.validator().validate(value):
                raise exceptions.FieldError(self.validator.validation_message)
            return value
        except exceptions.ChoiceError:
            raise
        except exceptions.FieldError:
            raise
        except ValueError:
            self.raise_type_error(value)

    def raise_type_error(self, value):
        raise ValueError("Value \'%s\' in columns %d does not match the expected type %s" %
                             (value, self.position + 1, self.__class__.field_name))


class IntegerField(Field):
    field_name = "Integer"

    def to_python(self, value):
        return int(value)


class BooleanField(Field):
    field_name = "Boolean"

    def default_is_true_method(self, value):
        return value.lower() == "true"

    def __init__(self, *args, **kwargs):
        if 'is_true' in kwargs:
            self.is_true_method = kwargs.pop('is_true')
        else:
            self.is_true_method = self.default_is_true_method
        super(BooleanField, self).__init__(*args, **kwargs)

    def to_python(self, value):
        return self.is_true_method(value)


class CharField(Field):
    field_name = "String"

    def to_python(self, value):
        return value


class DateField(Field):
    field_name = "Date"

    def __init__(self, *args, **kwargs):
        if 'format' in kwargs:
            self.format = kwargs.pop('format')
        else:
            self.format = "%d/%m/%Y"
        super(DateField, self).__init__(*args, **kwargs)

    def to_python(self, value):
        return datetime.strptime(value, self.format)


class DecimalField(Field):
    field_name = "A Decimal number"

    def to_python(self, value):
        return Decimal(value)


class FloatField(Field):
    field_name = "A float number"

    def to_python(self, value):
        return float(value)


class IgnoredField(Field):
    field_name = "Ignore the value"


class DjangoModelField(Field):
    field_name = "not defined"

    def __init__(self, *args, **kwargs):
        self.pk = kwargs.pop('pk', 'pk')
        if len(args) < 1:
            raise ValueError("You should provide a Model as the first argument.")
        self.model = args[0]
        try:
            if not issubclass(self.model, djangoModel):
                raise TypeError("The first argument should be a django model class.")
        except TypeError:
            raise TypeError("The first argument should be a django model class.")
        super(DjangoModelField, self).__init__(**kwargs)

    def to_python(self, value):
        try:
            return self.model.objects.get(**{self.pk: value})
        except ObjectDoesNotExist:
            raise exceptions.ForeignKeyFieldError("No match found for %s, %s:%s" % (self.model.__name__, self.pk, value), self.model.__name__, value)
        except MultipleObjectsReturned:
            raise exceptions.ForeignKeyFieldError("Multiple match found for %s" % self.model.__name__, self.model.__name__, value)


class ComposedKeyField(DjangoModelField):
    def to_python(self, value):
        try:
            return self.model.objects.get(**value)
        except ObjectDoesNotExist:
            raise exceptions.ForeignKeyFieldError("No match found for %s" % self.model.__name__, self.model.__name__, value)



from django.template import Context
from django.template.loader import get_template

from accountifie.toolkit.utils import get_bstrap_table


def daily_activity(dt):
    data_url = "/reporting/reports/AccountActivity/?col_tag=daily_%s&format=json" % dt.isoformat()
    row_defs = [{'data_field': 'label', 'value': 'Account', 'formatter': 'nameFormatter'},
                {'data_field': 'Yesterday', 'value': 'Yesterday', 'formatter': 'drillFormatter'},
                {'data_field': 'Change', 'value': 'Change', 'formatter': 'drillFormatter'},
                {'data_field': 'Today', 'value': 'Today', 'formatter': 'drillFormatter'},
            ]
    return get_bstrap_table(data_url, row_defs)


def tasks_list():
    data_url = "/api/celery/tasks_just_finished"
    row_defs = [{'data_field': 'task_name', 'value': 'Task Name', 'formatter': 'nameFormatter'},
                {'data_field': 'task_status', 'value': 'Status', 'formatter': 'nameFormatter'},
                {'data_field': 'date_done', 'value': 'Started', 'formatter': 'timeFormatter'},
                {'data_field': 'task_id', 'value': 'id', 'formatter': 'nameFormatter'},
                {'data_field': 'return_value', 'value': 'Return Value', 'formatter': 'nameFormatter'},
                {'data_field': 'download', 'value': 'Download', 'formatter': 'nameFormatter'},
            ]
    return get_bstrap_table(data_url, row_defs)


def metrics(label):
    data_url = "/api/reporting/metric/?label=%s" % label
    row_defs = [{'data_field': 'metric', 'value': 'Metric', 'formatter': 'nameFormatter'},
                {'data_field': 'label', 'value': 'Label', 'formatter': 'nameFormatter'},
                {'data_field': 'date', 'value': 'Date', 'formatter': 'dateFormatter'},
                {'data_field': 'balance', 'value': 'Value', 'formatter': 'nameFormatter'},
            ]
    return get_bstrap_table(data_url, row_defs)


def balance_trends(dt, acct_list=None, accts_path=None, company_id='EFIE'):
    if acct_list:
        data_url = "/reporting/api/balance_trends/?date=%s&acct_list=%s&company_id=%s" %( dt, '.'.join(acct_list), company_id)
    elif accts_path:
        data_url = "/reporting/api/balance_trends/?date=%s&accts_path=%s&company_id=%s" %( dt, accts_path, company_id)

    row_defs = [{'data_field': 'label', 'value': 'Date', 'formatter': 'nameFormatter'},
                {'data_field': 'M_2', 'value': '2 Months Ago', 'formatter': 'drillFormatter'},
                {'data_field': 'M_1', 'value': '1 Month Ago', 'formatter': 'drillFormatter'},
                {'data_field': 'M_0', 'value': 'This Month', 'formatter': 'drillFormatter'},
            ]
    return get_bstrap_table(data_url, row_defs)


def check_external_bals(dt, company_id='EFIE'):
    data_url = "/api/gl/check_external_bals/?date=%s&company_id=%s" % (dt.isoformat(), company_id)
    row_defs = [{'data_field': 'Account', 'value': 'Account', 'formatter': 'nameFormatter'},
                {'data_field': 'Internal', 'value': 'Internal', 'formatter': 'valueFormatter'},
                {'data_field': 'External', 'value': 'External', 'formatter': 'valueFormatter'},
                {'data_field': 'Diff', 'value': 'Diff', 'formatter': 'valueFormatter'},
            ]
    return get_bstrap_table(data_url, row_defs)

def external_bals_history(dt, company_id='EFIE', acct=''):
    data_url = "/api/gl/external_bals_history/?date=%s&company_id=%s&acct=%s" % (dt.isoformat(), company_id, acct)
    row_defs = [{'data_field': 'Date', 'value': 'Date', 'formatter': 'nameFormatter'},
                {'data_field': 'Internal', 'value': 'Internal', 'formatter': 'valueFormatter'},
                {'data_field': 'External', 'value': 'External', 'formatter': 'valueFormatter'},
                {'data_field': 'Diff', 'value': 'Diff', 'formatter': 'valueFormatter'},
            ]
    return get_bstrap_table(data_url, row_defs)


def snapshots():
    data_url = '/api/snapshot/glsnapshot/'
    row_defs = [{'data_field': 'id', 'value': 'ID', 'formatter': 'nameFormatter'},
                {'data_field': 'desc_link', 'value': 'Description', 'formatter': 'nameFormatter'},
                {'data_field': 'closing_date', 'value': 'Closing Date', 'formatter': 'nameFormatter'},
                {'data_field': 'snapped_at', 'value': 'Snapped At', 'formatter': 'nameFormatter'},
            ]
    return get_bstrap_table(data_url, row_defs)



"""
Strategy that performs transaction querying locally.
The tried and tested way of doing things, though not as performant as
query_manager_remote_strategy.py.
See query_manager_strategy.py for interface docs.
Use query_manager_strategy_factory.py to get an instance of this class.
"""

import dateutil.parser
import pandas as pd
import accountifie.gl.cache
from accountifie.common.api import api_func

import accountifie.toolkit.utils as utils

from decimal import Decimal, ROUND_HALF_UP
from functools import partial
from .query_manager_strategy import QueryManagerStrategy


class QueryManagerLocalStrategy(QueryManagerStrategy):

    def account_balances_for_dates(self, company_id, account_ids, dates, with_counterparties, excl_interco, excl_contra):
        date_indexed_account_balances = {}
        entries = self.__pd_balances_prep(company_id,
                                          account_ids,
                                          excl_interco=excl_interco,
                                          excl_contra=excl_contra,
                                          with_counterparties=with_counterparties)

        if entries is None:
            return {dt: {account: Decimal(0) for account in account_ids} for dt in dates}

        for dt in dates:
            start = dates[dt]['start']
            end = dates[dt]['end']

            sub_entries = self.__depreciation_calcs(start, end, entries)
            col = sub_entries[['account_id', 'amount']].groupby('account_id').sum()['amount']
            date_indexed_account_balances[dt] = col.to_dict()

        return date_indexed_account_balances

    def transactions(self, company_id, account_ids, from_date, to_date, chunk_frequency, with_counterparties, excl_interco, excl_contra):
        period_entries = []
        from_date = dateutil.parser.parse(from_date).date() if isinstance(from_date, str) else from_date
        to_date = dateutil.parser.parse(to_date).date() if isinstance(to_date, str) else to_date

        periods = [{'from': from_date, 'to': to_date}]
        if chunk_frequency == 'end-of-month':
            periods = self.monthly_chunk_periods(from_date, to_date)
        else:
            raise Exception('Unknown chunk frequency: %s' % chunk_frequency)

        cache = accountifie.gl.cache.get_cache(company_id)
        clean_end = lambda row: row['date_end'] if row['date_end'] else row['date']
        clean_id = lambda row: str(row['transaction_id'])
        for period in periods:
            start = period['from']
            end = period['to']

            entries = cache.get_gl_entries(account_ids, from_date=start, to_date=end)

            if entries is not None and not entries.empty:
                if with_counterparties:
                    entries = entries[entries['counterparty'] in with_counterparties]

                if excl_contra and not entries.empty:
                    for contra in excl_contra:
                        entries = entries[entries.contra_accts.map(lambda x: contra not in x.split(','))]

                if excl_interco and not entries.empty:
                    entries = entries[~entries.apply(self.__inter_co, axis=1)]

                if entries.empty:
                    break

                entries['transaction_id'] = entries.apply(clean_id, axis=1)
                entries['date_end'] = entries.apply(clean_end, axis=1)
                sub_entries = self.__depreciation_calcs(start, end, entries)

                sub_entries.reset_index(drop=False, inplace=True)
                sub_entries.rename(columns={'transaction_id': 'id'}, inplace=True)
                period_entries.append(sub_entries)

        if len(period_entries) == 0:
            return None
        else:
            all_entries = pd.concat(period_entries, ignore_index=True)
            return all_entries.filter(items=['date', 'id', 'comment', 'account_id', 'contra_accts', 'counterparty', 'amount']).to_dict()

    def __pd_balances_prep(self, company_id, account_ids, excl_contra=None, excl_interco=False, with_counterparties=None):
        if api_func('gl', 'company', company_id)['cmpy_type'] == 'CON':
            excl_interco = True

        cache = accountifie.gl.cache.get_cache(company_id)

        entries = cache.get_gl_entries(account_ids)

        if entries is None or entries.empty:
            return None

        if excl_contra:
            for contra in excl_contra:
                entries = entries[entries.contra_accts.map(lambda x: contra not in x.split(','))]

        if excl_interco:
            entries = entries[~entries.apply(self.__inter_co, axis=1)]

        if with_counterparties:
            entries = entries[entries['counterparty'] in with_counterparties]

        if entries.empty:
            return None

        clean_end = lambda row: row['date_end'] if row['date_end'] else row['date']
        entries['date_end'] = entries.apply(clean_end, axis=1)

        return entries


    """
    def create_gl_transactions(self, d2, lines, trans_id, bmo_id):

        
        tran = accountifie.gl.models.Transaction(**d2)
        tran.source_object = d2.pop('source_object')
        tran.bmo_id = bmo_id
        tran.save()

        for l in lines:
            if type(l['account']) in (str,):
                l['account'] = accountifie.gl.models.Account.objects.get(id=l['account'])
            if type(l['counterparty']) in (str,):
                l['counterparty'] = accountifie.gl.models.Counterparty.objects.get(id=l['counterparty'])
            tran.tranline_set.create(account=l['account'],
                                     amount=l['amount'],
                                     counterparty=l['counterparty'],
                                     tags=l['tags'])
    """

    def delete_bmo_transactions(self, company_id, bmo_id):
        logger.info('Deleting transactions: bmo ID=%s, Company=%s' % (company_id, bmo_id))
        for t in accountifie.gl.models.Transaction.objects.filter(bmo_id=bmo_id):
            t.delete()

    def __depreciation_calcs(self, start, end, entries):
        extend_comment = lambda row: row['comment'] + (' (prorated)' if row['deprec_factor'] < Decimal('1') else '')
        sub_entries = entries[(entries.date <= end) & (entries.date_end >= start)]
        _deprec_factor = partial(self.deprec_factor, start=start, end=end)

        if not sub_entries.empty:
            sub_entries.loc[:, 'deprec_factor'] = sub_entries.apply(_deprec_factor, axis=1)
            sub_entries.loc[:, 'comment'] = sub_entries.apply(extend_comment, axis=1)
            sub_entries.loc[:, 'amount'] *= sub_entries['deprec_factor']
            sub_entries.loc[:, 'amount'] = sub_entries['amount'].map(lambda x: x.quantize(Decimal('0.01'), rounding=ROUND_HALF_UP))
            sub_entries.loc[:, 'date_end'] = sub_entries.apply(lambda row: min(row['date_end'], end), axis=1)
        return sub_entries

    @staticmethod
    def monthly_chunk_periods(from_date, to_date):
        # dates should be from_date, to_date and the first and last days of every month included
        months = list(set(utils.monthrange(from_date, to_date)))
        dates = [(max(utils.start_of_month(x[1], x[0]), from_date), min(utils.end_of_month(x[1], x[0]), to_date)) for x in months]
        periods = [{'from': dt[0], 'to': dt[1]} for dt in dates]
        return periods

    @staticmethod
    def __inter_co(row):
        ext_accts = api_func('gl', 'externalaccounts')
        companies = [cmpy['id'] for cmpy in api_func('gl', 'company')]
        if row['account_id'] in ext_accts:
            return False
        if row['counterparty'] in companies:
            return True
        else:
            return False


    @staticmethod
    def deprec_factor(row, start, end):
        if row['date_end'] is None or row['date_end'] == row['date']:
            return Decimal('1')
        days_outside = Decimal(max((row['date_end'] - end).days, 0) + max((start - row['date']).days, 0))
        expense_period = Decimal((row['date_end'] - row['date']).days + 1)
        deprec_factor = Decimal('1')
        if expense_period > 0:
            deprec_factor -= days_outside / expense_period
        return deprec_factor


from django.contrib import admin
from .models import Variable, Config, Alias

class VariableAdmin(admin.ModelAdmin):
    ordering = ('key',)
    list_display = ('key', 'value')
    
class ConfigAdmin(admin.ModelAdmin):
    ordering = ('name',)
    list_display = ('name',)

class AliasAdmin(admin.ModelAdmin):
    ordering = ('name', 'context',)
    list_display = ('name', 'display_as', 'context',)
    list_filter = ('context',)

admin.site.register(Variable, VariableAdmin)
admin.site.register(Config, ConfigAdmin)
admin.site.register(Alias, AliasAdmin)



import pytz

from django.db import models
from django.utils.safestring import mark_safe

from accountifie.query.query_manager_strategy_factory import QueryManagerStrategyFactory
from accountifie.common.api import api_func

import logging

logger = logging.getLogger('default')

UTC = pytz.timezone('UTC')

class GLSnapshot(models.Model):
    snapped_at = models.DateTimeField(help_text="Local NY time")
    short_desc = models.CharField(max_length=100)
    comment = models.TextField()
    closing_date = models.DateField()


    def save(self):
        models.Model.save(self)
        logger.info('saving snapshot with snapshot time %s' % self.snapped_at)
        # now create snapshot
        for company_id in [c['id'] for c in api_func('gl', 'company') if c['cmpy_type']=='ALO']:
            fmt = '%Y-%m-%dT%H:%M:%SZ'
            snapshot_time = self.snapped_at.astimezone(UTC).strftime(fmt)
            QueryManagerStrategyFactory().get().take_snapshot(company_id, snapshot_time=snapshot_time)


    @property
    def desc_link(self):
      return mark_safe('<a href="/snapshot/glsnapshots/balances/%s/">%s' %( self.id, self.short_desc))


    @property
    def reconciliation(self):
        url = '/snapshot/glsnapshots/balances/%s/?date=%s' % (self.id, self.closing_date.isoformat())

        return mark_safe('<a href="%s">%s' % (url, 'Reconciliation'))
